# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

import scipy.linalg as spla

from .likelihood_function import LikelihoodFunction


class Gaussian(LikelihoodFunction):
    """

    """

    def log_likelihood(self, data, marginal):
        """
        Measure the log Likelihood

        :param data:
        :type data:
        :param marginal:
        :type marginal:
        :return:
        :rtype:
        """
        llikelihood = -np.sum(np.log(np.diag(marginal['l_matrix']))) - \
            0.5 * np.dot(
                    (data['Y'] - marginal['mean']).T,
                    marginal['alpha']
                )[0, 0] - \
            0.5 * marginal['l_matrix'].shape[0] * \
                np.log(2.0 * np.pi)

        return llikelihood

    def log_density(self, fy, var, mean):
        """
        Measure the log density

        :param fy:
        :type fy:
        :param var:
        :type var:
        :param mean:
        :type mean:
        :return:
        :rtype:
        """
        log_predictive_density = \
            -0.5 * np.log(var) - \
            np.power((fy - mean), 2) / (2.0 * var) - \
            0.5 * np.log(2 * np.pi)

        return log_predictive_density

    def dlog_likelihood_dtheta(self, data, marginal):
        """
        Measure the gradient log Likelihood

        :param data:
        :type data:
        :param marginal:
        :type marginal:
        :return:
        :rtype:
        """
        k_inv = spla.cho_solve(
            (marginal['l_matrix'], True),
            np.eye(marginal['l_matrix'].shape[0]))
        jacobian = np.outer(marginal['alpha'],
                            marginal['alpha']) - k_inv

        grad_llikelihood = []
        # Log amplitude gradient.
        _, dmu_dtheta = self.gp.mean_function.marginalize_mean(
            data['X'], dmu_dtheta_needed=True, trans=True)

        for dmu_dtheta_i in dmu_dtheta:
            grad_llikelihood.append(
                self._dlog_likelihood_dtheta(dmu_dtheta_i, jacobian))
        # Log amplitude gradient.
        _, dk_dtheta = self.gp.covariance_function.marginalize_covariance(
            data['X'], dk_dtheta_needed=True, trans=True)

        for dk_dtheta_i in dk_dtheta:
            grad_llikelihood.append(
                self._dlog_likelihood_dtheta(dk_dtheta_i, jacobian))

        return grad_llikelihood

    @staticmethod
    def _dlog_likelihood_dtheta(dk_dtheta, jacobian):
        """
        Measure the gradient of the log Likelihood

        :param jacobian:
        :return:
        """
        return 0.5 * np.trace(np.dot(jacobian, dk_dtheta))



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import time

import numpy as np

from .fitting_method import FittingMethod
from ..parameters.parameter import Parameter


class GridSearch(FittingMethod):
    """

    """

    def __init__(self, obj_fun, max_fun_call=500, group=None,
                 nested_fit_method=None):

        self.obj_fun = obj_fun
        self.group = Parameter.GRID_GROUP
        if group is not None:
            self.group = group
        self.nested_fit_method = nested_fit_method
        self.max_fun_call = max_fun_call

    def fit(self, model, folds, budget=None, verbose=False):
        """

        :param model:
        :type model:
        :param folds:
        :type folds:
        :param budget:
        :type budget:
        :param verbose:
        :type verbose:
        :return:
        :rtype:
        """
        if budget is None or self.max_fun_call < budget:
            max_fun_call = self.max_fun_call
        else:
            max_fun_call = budget

        start = time.time()

        log = {
            'fun_calls': 0,
            'improvements': 0,
            'restarts': 0,
            'time': 0,
            'best': {
                'params': model.get_param_values(
                    trans=False
                ),
                'value': np.inf,
                'fun_call': 0,
                'nested': None
            }
        }

        grid = model.get_grid(
            limit=max_fun_call,
            only_group=self.group,
            trans=False
        )
        i = 0

        if len(grid) < 1:
            grid = [[]]

        while log['fun_calls'] < max_fun_call and i < len(grid):
            current_params = grid[i]
            i += 1

            # run optimization
            value = np.inf
            model.set_param_values(
                current_params,
                only_group=self.group,
                trans=False
            )
            try:
                if self.nested_fit_method is not None:
                    nested_log = self.nested_fit_method.fit(
                        model,
                        folds,
                        budget=(max_fun_call - log['fun_calls']),
                        verbose=verbose
                    )
                    log['fun_calls'] += (nested_log['fun_calls'] - 1)
                    value = nested_log['best']['value']
                else:
                    value = self.obj_fun(
                        model=model,
                        folds=folds,
                        grad_needed=False
                    )
            except (AssertionError, np.linalg.linalg.LinAlgError) as ex:
                if verbose:
                    print(ex)

            log['fun_calls'] += 1
            log['restarts'] += 1
            if value < log['best']['value']:
                log['improvements'] += 1
                log['best']['params'] = model.get_param_values(
                    trans=False
                )
                log['best']['value'] = value
                log['best']['fun_call'] = log['fun_calls']
                if self.nested_fit_method is not None:
                    log['best']['nested'] = {
                        'name': self.nested_fit_method.__class__.__name__,
                        'fun_calls': nested_log['fun_calls'],
                        'improvements': nested_log['improvements'],
                        'restarts': nested_log['restarts'],
                        'time': nested_log['time'],
                        'best_fun_call': nested_log['best']['fun_call'],
                        'nested': nested_log['best']['nested']
                    }

        end = time.time()

        log['time'] = end - start

        assert log['best']['fun_call'], "No params were found"

        model.set_param_values(
            log['best']['params'],
            trans=False
        )

        return log



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import copy

import numpy as np
import scipy.linalg


class InferenceMethod(object):
    """

    """

    def __init__(self, cache_size=10):

        self.gp = None
        self.cache = {
            'size': cache_size,
            'cov_hashes': [],
            'data': [],
            'l_matrices': []
        }

    def __copy__(self):

        copyed_object = self.__class__()

        return copyed_object

    def set_gp(self, gp):
        """

        :param gp:
        :type gp:
        :return:
        :rtype:
        """
        self.gp = gp

    def marginalize_gp(self, data_x, data_y=None):
        """

        :param data_x:
        :type data_x:
        :param data_y:
        :type data_y:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def loocv(self, data_x, data_y=None):
        """

        :param data_x:
        :type data_x:
        :param data_y:
        :type data_y:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_l_matrix(self, data_x):
        """

        :param data_x:
        :type data_x:
        :return:
        :rtype:
        """
        cov_hash = hash(self.gp.covariance_function)
        param_hash = hash(tuple(self.gp.covariance_function.get_param_values()))
        cov_hash = hash((cov_hash, param_hash))

        isin_cache = np.isin(self.cache['cov_hashes'], cov_hash)
        l_matrix = None
        if np.any(isin_cache):
            match_i = np.argwhere(isin_cache)[-1, 0]
            cached_data = self.cache['data'][match_i]
            cached_l_matrix = self.cache['l_matrices'][match_i]

            data_size = min(cached_data.shape[0], data_x.shape[0])

            cached_data_hash = hash(cached_data[:data_size, :].tostring())
            data_x_hash = hash(data_x[:data_size, :].tostring())

            if cached_data_hash == data_x_hash:
                if cached_data.shape[0] == data_x.shape[0]:
                    return copy.deepcopy(cached_l_matrix)
                elif cached_data.shape[0] < data_x.shape[0]:
                    covariance = \
                        self.gp.covariance_function.marginalize_covariance(
                            data_x
                        )
                    l_matrix = InferenceMethod._update_chol(
                        covariance,
                        cached_l_matrix,
                        offset=0
                    )
                else:
                    covariance = \
                        self.gp.covariance_function.marginalize_covariance(
                            cached_data
                        )
                    l_matrix = InferenceMethod._downdate_chol(
                        covariance,
                        cached_l_matrix,
                        offset=0,
                        size=(cached_data.shape[0] - data_x.shape[0])
                    )

        if l_matrix is None:
            covariance = self.gp.covariance_function.marginalize_covariance(
                data_x
            )
            l_matrix = InferenceMethod._chol(covariance)

        if len(self.cache['cov_hashes']) > self.cache['size']:
            del self.cache['cov_hashes'][0]
            del self.cache['data'][0]
            del self.cache['l_matrices'][0]

        self.cache['cov_hashes'].append(cov_hash)
        self.cache['data'].append(data_x)
        self.cache['l_matrices'].append(l_matrix)

        return l_matrix

    @staticmethod
    def _chol(k_matrix):
        """
        Compute cholesky decomposition

        :param k_matrix:
        :type k_matrix:
        :return:
        :rtype:
        """
        InferenceMethod._check_k_matrix(k_matrix)

        # Solve cholesky decomposition
        l_matrix = InferenceMethod._jittered_chol(k_matrix)

        InferenceMethod._check_l_matrix(l_matrix, k_matrix)

        return l_matrix

    @staticmethod
    def _update_chol(k_matrix, l_matrix, offset=0):
        """

        :param k_matrix:
        :type k_matrix:
        :param l_matrix:
        :type l_matrix:
        :param offset:
        :type offset:
        :return:
        :rtype:
        """
        InferenceMethod._check_k_matrix(k_matrix)

        b_i = k_matrix.shape[0] - offset
        new_i = l_matrix.shape[0] - offset
        new_l_matrix = np.zeros(k_matrix.shape)
        new_l_matrix[:new_i, :new_i] = l_matrix[:new_i, :new_i]
        new_l_matrix[new_i:b_i:, :new_i] = scipy.linalg.solve_triangular(
            l_matrix[:new_i, :new_i],
            k_matrix[:new_i, new_i:b_i],
            lower=True,
            check_finite=False
        ).T

        new_l_matrix[b_i:, :new_i] = l_matrix[new_i:, :new_i]
        new_l_matrix[new_i:b_i, new_i:b_i] = InferenceMethod._jittered_chol(
            k_matrix[new_i:b_i, new_i:b_i] -
            np.dot(
                new_l_matrix[new_i:b_i:, :new_i],
                new_l_matrix[new_i:b_i:, :new_i].T
            )
        )
        new_l_matrix[b_i:, new_i:b_i] = scipy.linalg.solve_triangular(
            new_l_matrix[new_i:b_i, new_i:b_i],
            k_matrix[b_i:, new_i:b_i].T -
            np.dot(
                new_l_matrix[new_i:b_i:, :new_i],
                new_l_matrix[b_i:, :new_i].T
            ),
            lower=True,
            check_finite=False
        ).T
        new_l_matrix[b_i:, b_i:] = InferenceMethod._jittered_chol(
            np.dot(l_matrix[new_i:, new_i:], l_matrix[new_i:, new_i:].T) -
            np.dot(
                new_l_matrix[b_i:, new_i:b_i],
                new_l_matrix[b_i:, new_i:b_i].T
            )
        )

        InferenceMethod._check_l_matrix(new_l_matrix, k_matrix)

        return new_l_matrix

    @staticmethod
    def _downdate_chol(k_matrix, l_matrix, offset=0, size=0):
        """

        :param k_matrix:
        :type k_matrix:
        :param l_matrix:
        :type l_matrix:
        :param offset:
        :type offset:
        :param size:
        :type size:
        :return:
        :rtype:
        """
        InferenceMethod._check_k_matrix(k_matrix)

        b_i = k_matrix.shape[0] - offset
        new_i = b_i - size
        new_size = k_matrix.shape[0] - size
        new_l_matrix = np.zeros((new_size, new_size))
        new_l_matrix[:new_i, :new_i] = l_matrix[:new_i, :new_i]
        new_l_matrix[new_i:, new_i:] = InferenceMethod._jittered_chol(
            np.dot(l_matrix[b_i:, b_i:], l_matrix[b_i:, b_i:].T) +
            np.dot(l_matrix[b_i:, new_i:b_i], l_matrix[b_i:, new_i:b_i].T)
        )
        new_l_matrix[new_i:, :new_i] = l_matrix[b_i:, :new_i]

        new_k_matrix = np.vstack((
            np.hstack((k_matrix[:new_i, :new_i], k_matrix[:new_i, b_i:])),
            np.hstack((k_matrix[b_i:, :new_i], k_matrix[b_i:, b_i:]))
        ))

        InferenceMethod._check_l_matrix(new_l_matrix, new_k_matrix)

        return new_l_matrix

    @staticmethod
    def _check_k_matrix(k_matrix):
        """

        :param k_matrix:
        :type k_matrix:
        :return:
        :rtype:
        """

        # Empty
        if k_matrix.size == 0:
            return np.empty((0, 0))

        # Non finite values in covariance matrix
        if not np.all(np.isfinite(k_matrix)):
            raise np.linalg.LinAlgError("Non finite values in cov matrix")

        # Covariance matrix is not symmetric
        if np.max(np.abs(k_matrix - k_matrix.T)) > 1e-20:
            raise np.linalg.LinAlgError("Covariance matrix is not symmetric")

    @staticmethod
    def _jittered_chol(k_matrix):
        """

        :param k_matrix:
        :type k_matrix:
        :return:
        :rtype:
        """

        l_matrix = None
        jitter = 1e-30
        max_jitter = 1e-6
        k_corrected = k_matrix
        while l_matrix is None and jitter < max_jitter:
            l_matrix, error = scipy.linalg.lapack.dpotrf(
                np.ascontiguousarray(k_corrected),
                lower=1
            )
            if error != 0:
                l_matrix = None
                k_corrected = k_matrix + jitter * np.eye(k_matrix.shape[0])
                jitter *= 10

        if l_matrix is None:
            raise np.linalg.LinAlgError("Can't compute cholesky decomposition")

        return l_matrix

    @staticmethod
    def _check_l_matrix(l_matrix, k_matrix):
        """

        :param l_matrix:
        :type l_matrix:
        :param k_matrix:
        :type k_matrix:
        :return:
        :rtype:
        """

        # Non finite values in L matrix
        if not np.all(np.isfinite(l_matrix)):
            raise np.linalg.LinAlgError("Non finite values in L matrix")

        # Main diagonal of L not positive
        if np.min(np.diagonal(l_matrix)) <= 0.0:
            raise np.linalg.LinAlgError("Main diagonal of L not positive")

        # Errors in L matrix multiplication
        if np.max(np.abs(k_matrix - np.dot(l_matrix, l_matrix.T))) > 1e-4:
            raise np.linalg.LinAlgError("Errors in L matrix multiplication")
